##########################################################################
# NSAp - Copyright (C) CEA, 2013 - 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Module that provides tools to reorganize DICOM files.
"""


# System import
from __future__ import print_function
import os
import sys
import shutil
import string
import traceback

# Third party import
import progressbar
import dicom


def decode(attribute):
    """Decode DICOM attributes from ISO_IR 100.

    DICOM headers are routinely encoded with ISO_IR 100 which is
    equivalent to IS0 8859-1.

    We currently expect all our DICOM headers to be encoded using
    ISO_IR 100. In this context DICOM string attributes returned
    by pydicom are 8-bit strings encoded with ISO_IR 100.

    Parameters
    ----------
    attribute  : str
        The 8-bit string to decode from ISO_IR 100.

    Returns
    -------
    unicode
        The decoded string.

    """
    return attribute.decode("latin_1")


_ILLEGAL_CHARACTERS = u"\\/:*?'<>|_ \t\r\n\0[],;"
_CLEANUP_TABLE = dict((ord(char), u"-") for char in _ILLEGAL_CHARACTERS)


def cleanup(attribute):
    """Get rid of illegal characters in DICOM attributes.

    Replace characters that are illegal in pathnames.
    - Windows reserved characters
    - '_' since it is reserved by Brainvisa
    - spaces, tab, newline and null character

    Parameters
    ----------
    attribute  : unicode
        Decoded string.

    Returns
    -------
    unicode
        String with illegal characters replaced.

    """
    return attribute.translate(_CLEANUP_TABLE)


def split_series(dicom_dir, outdir, skip_non_dicom_files=False,
                 check_session=True, check_encoding=True):
    """ Split all the folder Dicom files by series in different folders.

    Dicom files are searched recursively in the input folder and all files
    are expected to be Dicom files.

    Expect to split files from a single session.

    Parameters
    ----------
    dicom_dir: str (mandatory)
        a folder containing Dicom files to organize by series.
    outdir: str (mandatory)
        the destination folder.
    skip_non_dicom_files: bool (optional, default False)
        if True skip non DICOM files, otherwise raise an error.
    check_session: bool (optional, default True)
        if True check if the DICOM files are in the same session and split
        files by sequences (raise an error if it is not the case), otherwise
        simply rename the DICOM files using the SOP instance UID.
    check_encoding: bool (optional, default True)
        if True check if the DICOM files encoding (expect ISO_IR 100). If the
        file is not encoded properly, the file is not considered.
    """
    # Read the incoming directory:
    # process each file in this directory and its sub-directories
    # expect each file to be a DICOM file
    to_treat_dicom = []
    for root, dirs, files in os.walk(dicom_dir):
        to_treat_dicom.extend([
            os.path.join(root, basename) for basename in files])

    # Go through each file: expected to be in Dicom format
    acquisition_datetime = None
    with progressbar.ProgressBar(max_value=len(to_treat_dicom),
                                 redirect_stdout=True) as bar:
        for cnt, dicom_file in enumerate(to_treat_dicom):
            _split(dicom_file, acquisition_datetime, outdir,
                   skip_non_dicom_files, check_session, check_encoding)
            bar.update(cnt)


def safe_run(func):
    """ Decorator that print the function args and kwargs before raising an
    exception.
    """
    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            print("-" * 50)
            print(args)
            print(kwargs)
            print("-" * 50)
            raise e
    return func_wrapper


@safe_run
def _split(dicom_file, acquisition_datetime, outdir, skip_non_dicom_files,
           check_session, check_encoding):
    """ Split function, see 'split_series' for more information.
    """
    # Get the time of last modification
    mtime = os.path.getmtime(dicom_file)

    # Read DICOM dataset
    try:
        dataset = dicom.read_file(dicom_file)
    except:
        if skip_non_dicom_files:
            return
        traceback.print_exc(file=sys.stdout)
        raise ValueError(
            "'{0}' is not a valid DICOM file.".format(dicom_file))

    # Find character encoding of DICOM attributes:
    # we currently expect encoding to be ISO_IR 100
    if check_encoding:
        if (0x0008, 0x0005) in dataset:
            SpecificCharacterSet = dataset[0x0008, 0x0005].value
            if SpecificCharacterSet != "ISO_IR 100":
                print("'{0}' file encoding is not ISO_IR 100 as "
                      "expected.".format(dicom_file))
                return
        else:
            print("Can't check encoding of '{0}', missing "
                  "(0x0008, 0x0005) tag.".format(dicom_file))

    # Process other DICOM attributes:
    # decode strings assuming 'ISO_IR 100'
    SeriesDescription = None
    if (0x0008, 0x0018) not in dataset:
        if skip_non_dicom_files:
            return
        raise ValueError(
            "'{0}' does not contain a SOPInstanceUID.".format(
                dicom_file))
    SOPInstanceUID = dataset[0x0008, 0x0018].value
    if (0x0008, 0x103e) in dataset:
        SeriesDescription = cleanup(
            decode(dataset[0x0008, 0x103e].value))
    if check_session:
        SeriesNumber = dataset[0x0020, 0x0011].value
        if (0x0018, 0x0081) in dataset:
            EchoTime = dataset[0x0018, 0x0081].value
        else:
            EchoTime = "NA"

    # Check the session time
    if check_session:
        current_acquisition_datetime = (dataset[0x0008, 0x0020].value +
                                        dataset[0x0008, 0x0030].value)
        if acquisition_datetime is None:
            acquisition_datetime = current_acquisition_datetime
        elif acquisition_datetime != current_acquisition_datetime:
            print(
                "Two sessions detected in the input folder '{0}': {1} "
                "- {2}.".format(dicom_dir, acquisition_datetime,
                                current_acquisition_datetime))
            if SeriesDescription:
                SeriesDescription += "_{0}".format(
                    current_acquisition_datetime)
            else:
                SeriesDescription = current_acquisition_datetime

    # Build the full path to the outgoing directory:
    # we assume that there is only one session
    if check_session:
        if SeriesDescription:
            serie_name = (SeriesDescription + "_" + str(EchoTime) +
                          "_" + str(SeriesNumber).rjust(6, "0"))
        else:
            serie_name = (str(EchoTime) + "_" +
                          str(SeriesNumber).rjust(6, "0"))
    else:
        serie_name = "all_dicoms"
    output_dicom_dir = os.path.join(outdir, serie_name)

    # Check that the destination folder exists
    if not os.path.isdir(output_dicom_dir):
        os.mkdir(output_dicom_dir)

    # Build a new name for the DICOM file
    output_dicom_file = os.path.join(output_dicom_dir,
                                     SOPInstanceUID + '.dcm')

    # Copy DICOM file:
    # handle case where outgoing file already exists
    if os.path.exists(output_dicom_file):

        # Compare modification time and keep the most recent file
        if os.path.getmtime(output_dicom_file) < mtime:
            shutil.copy2(dicom_file, output_dicom_file)

    # file does not exists and can be copied
    else:
        shutil.copy2(dicom_file, output_dicom_file)



##########################################################################
# NSAp - Copyright (C) CEA, 2013 - 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Parse private Siemens CSA2 header field.
"""

# System import
import os
import ast
import dicom
import struct


def get_siemens_csa2_header(dataset_or_dcmpath):
    """ Return a dictionary with the Siemens CSA2 Header.

    Parameters
    ----------
    dataset_or_dcmpath: dataset or str (mandatory)
        a pydicom dataset structure or a path to a valid Dicom file.

    Returns
    -------
    content: dict
        a dictionnary containing the Siemens CSA2 header as (tag, values)
        items.
    """
    # Deal with input parameters
    if isinstance(dataset_or_dcmpath, str):
        if not os.path.isfile(dataset_or_dcmpath):
            raise ValueError("'{0}' is not a valid Dicom file.".format(
                dataset_or_dcmpath))
        dataset = dicom.read_file(dataset_or_dcmpath, force=True)
    else:
        if not isinstance(dataset_or_dcmpath, dicom.dataset.Dataset):
            raise ValueError("'{0}' is not a 'pydicom' Dataset.".format(
                dataset_or_dcmpath))
        dataset = dataset_or_dcmpath

    # Get the csa2 header representation
    content = None
    for tag in [(0x0029, 0x1010), (0x0029, 0x1020), (0x0029, 0x1210),
                (0x0029, 0x1110)]:
        if tag in dataset:
            content = parse_csa2(dataset[tag])
            break

    return content


def parse_csa2(csa):
    """ Return a dictionary with the Siemens CSA2 Header.

    Siemens CSA Header tag is (0x0029, 0x1020).
    See also http://nipy.org/nibabel/dicom/siemens_csa.html.

    Parameters
    ----------
    csa: binary (mandatory)
        the Siemens CSA2 Header.

    Returns
    -------
    content: dict
        a dictionnary containing the Siemens CSA2 header as (tag, values)
        items.
    """
    format = ("<"   # Little-endian
              "4s"  # SV10
              "4s"  # \x04\x03\x02\x01
              "I"   # Number of items
              "I"   # Unknown
              )
    size = struct.calcsize(format)

    version, _, number_of_elements, _ = struct.unpack(format, csa[:size])

    start = size
    content = {}
    for _ in range(number_of_elements):
        (name, items), size = parse_element(csa, start)
        content[name] = items
        start += size

    return content


def parse_element(csa, start):
    """ Return a pair (name, items), total_size.

    See also http://nipy.org/nibabel/dicom/siemens_csa.html.
    """
    format = ("<"    # Little endian
              "64s"  # Name
              "I"    # VM
              "2s"   # VR
              "2s"   # Unknown (end of VR ?)
              "I"    # Syngo datatype
              "I"    # Number of items
              "I"    # Unknown
              )
    size = struct.calcsize(format)

    name, vm, vr, _, syngo_datatype, number_of_items, _ = struct.unpack(
        format, csa[start:start + size])
    name = name.split("\x00")[0]

    total_size = size
    start += size
    items = []
    for i in range(number_of_items):
        item, size = parse_item(csa, start)
        if i < vm:
            if vr in ["DS", "FL", "FD"]:
                item = float(item[:-1])
            elif vr in ["IS", "SS", "US", "SL", "UL"]:
                item = int(item[:-1])
            else:
                try:
                    item = ast.literal_eval(item)
                except:
                    pass
            items.append(item)
        start += size
        total_size += size

    return (name, items), total_size


def parse_item(csa, start):
    """ Return a pair content, size.

    See also http://nipy.org/nibabel/dicom/siemens_csa.html.
    """
    format = ("<"   # Little endian
              "4I"  # Length
              )
    header_size = struct.calcsize(format)

    length = struct.unpack(format, csa[start:start + header_size])

    format = ("<"     # Little endian
              "{0}s"  # Content
              "{1}s"  # Padding (?)
              ).format(length[1], (4 - length[1] % 4) % 4)
    content_size = struct.calcsize(format)
    content, padding = struct.unpack(
        format,
        csa[start + header_size:start + header_size + content_size])

    return content, header_size + content_size



##########################################################################
# NSAp - Copyright (C) CEA, 2013 - 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Module that provides image snap tools.
"""


# System import
import numpy
import nibabel
import os
import math
import time
import numpy as np
import matplotlib
matplotlib.use("Agg")
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import FigureCanvasPdf as FigureCanvas


def mosaic(impath, outdir, strategy="average", indices=None, title=None,
           overlay=None, overlay_alpha=None, basename=None, ext=".pdf"):
    """ Create a snap of an input 3D or 4D image.

    If a 4D image is provided, select the 'index'th element or create an
    average volume.

    Parameters
    ----------
    impath: str
        the path to the image to slice.
    outdir: str
        the destination folder.
    strategy: str, default 'average'
        in the case of 4d image the slice strategy: 'pick' or 'average'.
    indices: list of int, default None
        in the case of 4d image the indices of the volumes to average.
    title: str, default None
        the mosaic title.
    overlay: str, default None
        the path to an mask to be overlayed.
    overlay_alpha: float, default None
        fix the overlay alpha value (0-1).
    basename: str, default None
        filename without extension of output file. By default None, means
        the input file basename will be used.
    ext: str, default '.png'
        snapshot extension, used to specify the output format.

    Returns
    -------
    snap: str
        a 'pdf' snap of the desired volume.
    """
    # Check the strategy
    if strategy not in ["pick", "average"]:
        raise ValueError("Uknown '{0}' 4d strategy.".format(strategy))

    # Load the input image and apply the 4d strategy if necessary
    array = nibabel.load(impath).get_data()
    if len(array.dtype) > 0:
        array = numpy.asarray(array.tolist())
    shape = array.shape

    if array.ndim < 3 or array.ndim > 4:
        raise ValueError("'{0}' is not a 3d or 4d image.".format(impath))
    if array.ndim == 4 and strategy == "average":
        array = numpy.mean(array, axis=3)
    if array.ndim == 4 and strategy == "pick":
        try:
            array = numpy.mean(array[..., indices], axis=3)
        except:
            raise ValueError("Can't pick volume indices '{0}' in array of "
                             "shape '{1}'.".format(indices, array.shape))

    # Create the snap with qap
    if basename is None:
        basename = os.path.basename(impath).split(".")[0]
    if title is None:
        title = basename
    title += ": shape {0}".format(shape)
    snap = os.path.join(outdir, basename + ext)
    fig = plot_mosaic(array, title=title, overlay_mask=overlay,
                      overlay_alpha=overlay_alpha)
    fig.savefig(snap, dpi=300)

    return snap


def plot_mosaic(nifti_file, title=None, overlay_mask=None, overlay_alpha=None,
                figsize=(11.7, 8.3)):
    """ From the qap module.
    """
    from pylab import cm

    if isinstance(nifti_file, str):
        nii = nibabel.load(nifti_file)
        mean_data = nii.get_data()
    else:
        mean_data = nifti_file

    z_vals = np.array(range(0, mean_data.shape[2]))

    # Crop inferior and posterior
    rem = 3
    mean_data = mean_data[..., rem:-rem]
    z_vals = z_vals[rem:-rem]
    # Reduce the number of slices shown
    if mean_data.shape[2] > 70:
        # Discard one every two slices
        mean_data = mean_data[..., ::2]
        z_vals = z_vals[::2]

    n_images = mean_data.shape[2]
    row, col = _calc_rows_columns(figsize[0] / figsize[1], n_images)

    if overlay_mask is not None:
        overlay_data = nibabel.load(overlay_mask).get_data()
        # Crop inferior and posterior
        overlay_data = overlay_data[..., rem:-rem]
        # Reduce the number of slices shown
        if mean_data.shape[2] > 70:
            # Discard one every two slices
            overlay_data = overlay_data[..., ::2]
        vmin = numpy.nanmin(overlay_data)
        vmax = numpy.nanmax(overlay_data)
        overlay_mask = np.logical_not(np.isnan(mean_data))
        if vmin != 0:
            vmin = np.percentile(overlay_data[overlay_mask], 1.0)
        if vmax != 1:
            vmax = np.percentile(overlay_data[overlay_mask], 99.0)

    # create figures
    fig = plt.Figure(figsize=figsize)
    FigureCanvas(fig)

    fig.subplots_adjust(top=0.85)
    for image, z_val in enumerate(z_vals):
        ax = fig.add_subplot(row, col, image + 1)
        data_mask = np.logical_not(np.isnan(mean_data))
        if overlay_mask is not None:
            ax.set_rasterized(True)

        ax.imshow(np.fliplr(mean_data[:, :, image].T), vmin=np.percentile(
            mean_data[data_mask], 0.5),
            vmax=np.percentile(mean_data[data_mask], 99.5),
            cmap=cm.Greys_r, interpolation='nearest', origin='lower')

        if overlay_mask is not None:
            cmap = cm.Reds
            cmap._init()
            if overlay_alpha is None:
                alphas = np.linspace(0, 0.75, cmap.N + 3)
                cmap._lut[:, -1] = alphas
            else:
                cmap._lut[:, -1] = overlay_alpha
            ax.imshow(np.fliplr(overlay_data[:, :, image].T), vmin=vmin,
                      vmax=vmax, cmap=cmap, interpolation='nearest',
                      origin='lower')

        ax.annotate(
            str(z_val), xy=(.95, .015), xycoords='axes fraction',
            fontsize=10, color='white', horizontalalignment='right',
            verticalalignment='bottom')

        ax.axis('off')

    fig.subplots_adjust(
        left=0.05, right=0.95, bottom=0.05, top=0.95, wspace=0.01, hspace=0.1)

    if not title:
        _, title = os.path.split(nifti_file)
        title += " (last modified: %s)" % time.ctime(
            os.path.getmtime(nifti_file))
    fig.suptitle(title, fontsize='10')
    return fig


def _calc_rows_columns(ratio, n_images):
    """ From the qap module.
    """
    rows = 1
    for _ in range(100):
        columns = math.floor(ratio * rows)
        total = rows * columns
        if total > n_images:
            break

        columns = math.ceil(ratio * rows)
        total = rows * columns
        if total > n_images:
            break
        rows += 1
    return rows, columns



##########################################################################
# NSAp - Copyright (C) CEA, 2013 - 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Dcm2Nii package excaptions.
"""


class Dcm2NiiError(Exception):
    """ Base exception type for the package.
    """
    def __init__(self, message):
        super(Dcm2NiiError, self).__init__(message)


class Dcm2NiiRuntimeError(Dcm2NiiError):
    """ Error thrown when call to the Dcm2Nii software failed.
    """
    def __init__(self, algorithm_name, parameters, error=None):
        message = (
            "Dcm2Nii call for '{0}' failed, with parameters: '{1}'. Error:: "
            "{2}.".format(algorithm_name, parameters, error))
        super(Dcm2NiiRuntimeError, self).__init__(message)


class Dcm2NiiConfigurationError(Dcm2NiiError):
    """ Error thrown when call to the Dcm2Nii software failed.
    """
    def __init__(self, command_name):
        message = "Dcm2Nii command '{0}' not found.".format(command_name)
        super(Dcm2NiiConfigurationError, self).__init__(message)


class Dcm2NiiResultError(Dcm2NiiError):
    """ Error thrown when the Dcm2Nii software has returned a strange result.
    """
    def __init__(self, command):
        message = ("Dcm2Nii command '{0}' may have returned a strange "
                   "result.".format(command))
        super(FSLResultError, self).__init__(message)



##########################################################################
# NSAP - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Module that provides common tools to anonimize DICOM files. Note that this
module is experimental.
"""



##########################################################################
# NSAp - Copyright (C) CEA, 2017
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Defacing with the Marcus, D. 'mask_face' command.
"""

# System import
import os
import shutil
import nibabel
import warnings

# Pyconnectome
try:
    from pyconnectome import DEFAULT_FSL_PATH
    from pyconnectome.wrapper import FSLWrapper
except:
    DEFAULT_FSL_PATH = ""
    warnings.warn("PyConnectome is not installed.")


def deface(input_files, outdir, matlab_mcr, reference_file=None,
           mask_ears=True, verbose=0, rm_workspace=False,
           fsl_sh=DEFAULT_FSL_PATH):
    """ Deface MRI head images using the Marcus, D. 'mask_face' command.

    This code is standalone and does not need matlab licence. You just need
    MathWorks MCR software installed on your computer.

    Parameters
    ----------
    input_files: list of str
        Input MRI head images to be defaced.
    outdir: str
        The output folder.
    matlab_mcr:str
        The MATLAB MCR directory for standalone applications.
    reference_file: str (optional, default None)
        The image that must be used as reference if more than one image
        have been supplied as input.
    mask_ears: bool (optional, default True)
        If activated, masks ears.
    verbose: int (optional, default 0)
        The verbosity level.
    rm_workspace: bool (optional, default False)
        If activated, keep the defacing workspace (require more disk space).
    fsl_sh: str (optional, default DEFAULT_FSL_PATH)
        The FSL configuration file.

    Returns
    -------
    deface_files: list of str
        The defaced input MRI head images.
    snap_files: list of str
        The corresponding snaps that can be used to check the defacing result.
    """
    # Check input parameters
    outdir = os.path.abspath(outdir)
    if len(input_files) == 0:
        raise ValueError("You must specify at least one image.")
    elif len(input_files) > 1:
        if reference_file is None or reference_file not in input_files:
            raise ValueError("If more than one image is specified, you must "
                             "also specify one of them as reference. The "
                             "reference image will be used for spatial "
                             "co-registration with an atlas, and others will "
                             "use the reference facial mask.")

    # Convert in analyse format
    pair_files = []
    basenames = []
    affines = []
    for fp in input_files:
        fp = os.path.abspath(fp)
        basenames.append(os.path.basename(fp).split(".")[0])
        pair_filepath = os.path.join(outdir, "{0}.img".format(basenames[-1]))
        im = nibabel.load(fp)
        affines.append(im.affine)
        pair_im = nibabel.Nifti1Pair(im.get_data(), im.affine)
        pair_im.to_filename(pair_filepath)
        pair_files.append(pair_filepath)

    # Define the command
    cmd = ["mask_face", ",".join([fp[:-len(".img")] for fp in pair_files])]
    cmd += ["-a"]
    cmd += ["-v", "{0}".format(verbose)]
    if reference_file is not None:
        ref_index = input_files.index(reference_file)
        cmd += ["-r", pair_files[ref_index]]
    if mask_ears:
        cmd += ["-e", "1"]

    # Call defacing
    deface_env = os.environ
    deface_env["MASKFACE_MCR"] = matlab_mcr
    wrapper = FSLWrapper(
        cmd, shfile=fsl_sh, env=deface_env)
    wrapper(cwdir=outdir)

    # Deal with outputs: nifti + move
    deface_files = []
    snap_files = []
    wdir = os.path.join(outdir, "maskface")
    for fp, basename, aff in zip(pair_files, basenames, affines):
        deface_pair_file = os.path.join(
            wdir, "{0}_full_normfilter.hdr".format(basename))
        nii_file = os.path.join(
            outdir, "{0}_full_normfilter.nii.gz".format(basename))
        im = nibabel.load(deface_pair_file)
        nii_im = nibabel.Nifti1Image(im.get_data(), aff)
        nibabel.save(nii_im, nii_file)
        del im, nii_im
        deface_files.append(nii_file)
        snap_files.append(
            os.path.join(outdir, "{0}_normfilter.png".format(basename)))
        shutil.copy2(
            os.path.join(wdir, "{0}_normfilter.png".format(basename)),
            snap_files[-1])
        os.remove(fp)
        os.remove(fp.replace(".img", ".hdr"))
    if rm_workspace:
        shutil.rmtree(wdir)
    else:
        shutil.move(wdir, os.path.join(outdir, "workspace"))

    return deface_files, snap_files



##########################################################################
# NSAp - Copyright (C) CEA, 2013 - 2018
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Diverse RTstruct Dicom reading functions to extract precise information.
"""

# System import
from __future__ import print_function
import dicom
import os
from pprint import pprint

# Third party import
import numpy as np
import nibabel
from pyfreesurfer.utils.surftools import apply_affine_on_mesh
from pyconnectome.utils.reorient import swap_affine


def regions_of_interest(rtstruct_file):
    """ Return list of all structure names.

    Parameters
    ----------
    rtstruct_file: str
        the RTstruct Dicom files.

    Returns
    -------
    rois: dict
        the ROI information: name, number of point, number of slices.
    """
    rois = {}
    dataset = dicom.read_file(rtstruct_file)
    rois_struct = dataset.StructureSetROISequence
    names = [rois_struct[idx].ROIName for idx in range(len(rois_struct))]
    for cnt, name in enumerate(names):
        slices = dataset.ROIContours[cnt].Contours
        num_slices = len(slices)
        if slices[0].ContourGeometricType == "CLOSED_PLANAR":
            num_pts = [slices[idx].NumberOfContourPoints
                       for idx in range(num_slices)]
            num_pts = np.array(num_pts).astype(int)
            num_pts = np.sum(num_pts)
            rois.update({
                name: {
                    "num_slices": num_slices,
                    "num_pts": num_pts}
            })
    return rois


def points_of_interest(rtstruct_file):
    """ Find POIs return as dict and display names and coordinates.

    Parameters
    ----------
    rtstruct_file: str
        the RTstruct dicom files.


    Returns
    -------
    rois: dict
        the ROI information: name, points in physical space, method.
    sop_instance_uids: list of str
        the associated volume slice SOp instance UIDs.
    institution: str
        the name of the institution in charge of the segmentation.
    operator: str
        the name of the person in charge of the segmentation.
    """
    rois = {}
    dataset = dicom.read_file(rtstruct_file)
    operator = dataset.OperatorsName
    institution = dataset.InstitutionName
    rois_struct = dataset.StructureSetROISequence
    names = [rois_struct[idx].ROIName for idx in range(len(rois_struct))]
    for cnt, name in enumerate(names):
        cntr = dataset.ROIContours[cnt].Contours[0]
        method = rois_struct[cnt].ROIGenerationAlgorithm
        num_pts = cntr.NumberOfContourPoints
        if cntr.ContourGeometricType == "CLOSED_PLANAR":
            try:
                frames = []
                for cntr_seq in cntr.ContourImageSequence:
                    frames.append(int(cntr_seq.RefdFrameNumber))
            except:
                frames = None
            pt_loc = np.asarray([float(xyz) for xyz in cntr.ContourData])
            pt_loc = pt_loc.reshape(-1, 3)
            rois.update({
                name: (pt_loc.tolist(), method, frames)
            })
    sop_instance_uids = []
    for slice_struct in (dataset.ReferencedFrameOfReferenceSequence[0].
                         RTReferencedStudySequence[0].
                         RTReferencedSeriesSequence[0].ContourImageSequence):
        sop_instance_uids.append(slice_struct.ReferencedSOPInstanceUID)
    return rois, sop_instance_uids, institution, operator


def generate_masks(rois, ref_file, outdir, axes="RAS", fname="mask"):
    """ Genrate Nifti masks from RTstruct data.

    Parameters
    ----------
    rois: dict
        the ROI information: name, points in physical space, method.
    ref_file: str
        the Nifti mask reference file.
    outdir: str
        the destination folder
    axes: str, default 'RAS'
        the ROIs points orientation axes.
    fname: str, default 'mask'
        the name of the generated mask file.

    Returns
    -------
    mask_file: str
        the generated masks from the RTstruct: a 4D volume.
    """
    ref_im = nibabel.load(ref_file)
    ref_data = ref_im.get_data()
    shape = ref_data.shape
    reorient_aff = swap_affine(axes)
    affine = np.dot(reorient_aff, ref_im.affine)
    inv_aff = np.linalg.inv(affine)
    mask = None
    for name, (pt_loc, method, frames) in rois.items():
        cmask = np.zeros(shape + (1, ), dtype=int)
        pt_loc = np.asarray(pt_loc)
        pt_loc_vox = np.round(apply_affine_on_mesh(
            pt_loc, inv_aff)).astype(int)
        for pt in pt_loc_vox:
            cmask[tuple(pt)] = 1
        if mask is None:
            mask = cmask
        else:
            mask = np.concatenate((mask, cmask), axis=-1)
    mask_im = nibabel.Nifti1Image(mask, ref_im.affine)
    mask_file = os.path.join(outdir, "{0}.nii.gz".format(fname))
    nibabel.save(mask_im, mask_file)
    return mask_file



##########################################################################
# NSAP - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Module that provides tools to anonymize DICOM files.
"""

# System import
import os
import json
import re

# Third party import
import progressbar
import dicom

# Dcmio import
from .callbacks import callback_private
from .callbacks import callback_xxxx
from .callbacks import ANON_LOG
from .callbacks import MANUFACTURER
from .callbacks import PRIVATE_DEIDENTIFY
from .callbacks import CALLBACKS
from .callbacks import TAGS
from .utils import add_dataelement
from .utils import replace_by
from .utils import repr_dataelement


def anonymize_dicomdir(inputdir, outdir, write_logs=True,
                       use_dicom_names=False, remove_all_private_tags=False):
    """ Anonymize all DICOM files of the input directory.

    Parameters
    ----------
    inputdir: str (mandatory)
        A folder that contains only DICOM files to be anonymized.
    outdir: str (mandatory)
        The anonimized DICOM files folder.
    write_logs: bool (optional, default True)
        If True write the anonimization logs.
    use_dicom_names: bool (optional, default False)
        If set use the input DICOM file names as anonymized DICOM file names
        (removing the file extension), otherwise just use a random numbering.
    remove_all_private_tags: bool (optional, default False)
        If set remove all the private tags in the DICOM files, otherwise
        apply the mapping available in the 'private_deidentify' file.

    Returns
    -------
    dcmfiles: str
        The anonimized DICOM files.
    logfiles: list
        The anonimization log files.
    """
    # Load the first dataset
    # Do not consider hidden files
    input_dicoms = [os.path.join(inputdir, fname)
                    for fname in os.listdir(inputdir)
                    if not fname.startswith(".")]
    dataset = dicom.read_file(input_dicoms[0], force=True)

    # Load the tags to anonymize
    filedir = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(filedir, "deidentify.json"), "r") as open_file:
        anon_tags = json.load(open_file)[1:]

    # Set up the desired callbacks and tags to be anonymized
    # Iterate over all the tag to anonymize according to PS 3.15-2008 and
    # supplement 142
    for tag_item in anon_tags:
        tag_repr = tag_item["Tag"][1:-1]
        action = tag_item["Basic Profile"]
        group, element = tag_repr.split(",", 1)

        # Deal with special tags
        if "xx" in group or "xx" in element:
            pattern = re.compile(tag_repr.replace("x", "[0-9A-Fa-f]"))
            CALLBACKS[tag_repr] = [pattern, callback_xxxx]

        # Deal with private tags
        elif "gggg" in group:
            if (0x0008, 0x0070) in dataset:
                MANUFACTURER.append(dataset[0x0008, 0x0070].value)
            if len(MANUFACTURER) > 0:
                CALLBACKS[tag_repr] = [None, callback_private]
            else:
                raise Exception(
                    "The '(0008,0070)' manufacturer tag is not specified and "
                    "is required to anonymize private tags.")

        # Deal with standard tags
        else:
            TAGS[tag_repr] = (int(group, 16), int(element, 16)), action

    # Now compile the diffusion private tags patterns
    filedir = os.path.dirname(os.path.realpath(__file__))
    if remove_all_private_tags:
        private_anons = {}
    else:
        with open(os.path.join(filedir, "private_deidentify.json"),
                  "r") as open_file:
            private_anons = json.load(open_file)
    for key, values in private_anons.items():
        for value in values:
            pattern = re.compile(value["Tag"].replace("x", "[0-9A-Fa-f]"))
            PRIVATE_DEIDENTIFY.setdefault(key, []).append(pattern)

    # Process all DICOM files
    dcmfiles = []
    logfiles = []
    with progressbar.ProgressBar(max_value=len(input_dicoms),
                                 redirect_stdout=True) as bar:
        for cnt, input_dicom in enumerate(input_dicoms):
            if use_dicom_names:
                otuname = os.path.basename(input_dicom).rsplit(".", 1)[0]
            else:
                otuname = str(cnt)
            output_dicom, output_log = anonymize_dicomfile(
                input_dicom, outdir, outname=otuname, write_log=write_logs)
            dcmfiles.append(output_dicom)
            logfiles.append(output_log)
            bar.update(cnt)

    return dcmfiles, logfiles


def anonymize_dicomfile(input_dicom, outdir, outname=None, write_log=True):
    """ Anonymize DICOMs

    According to PS 3.15-2008, basic application level de-indentification of
    a DICOM file requires replacing the values of a set of data elements and
    the supplement 142 on the clinical trial de-identification profiles. All
    the anonymized fields are replaced and may be set in future into the
    encrypted attributes sequence (0400,0550). In the standard it is not
    required that the Encrypted Attributes Dataset be created; indeed, there
    may be circumstances where the Dataset is expected to be archived long
    enough that any contemporary encryption technology may be inadequate
    to provide long term protection against unauthorized recovery of
    identification.

    The applied de-identification methods are:

    * 113100 - Basic Application Confidentiality Profil
    * 113103 - Clean Graphics Option
    * 113109 - Retain Device Identity Option

    To do:

    * 113102 - Clean Recognizable Visual Features Option

    What does this function:

    * All fields specified in the 'anon_tags' structure are replaced with
      the 'anon_value'.
    * To remove all private tags a group 'gggg' has to be specified in the
      'anon_tags' structure.
    * A log is save with all the anonymization operations.

    Parameters
    ----------
    input_dicom: str (mandatory)
        a dicom file path to be processed.
    outdir: str (mandatory)
        the directory where the anonymized DICOM file and a Json file
        containing a dictionary of the modified tags are generated.
    outname: str (optional, default None)
        the generated files base names, if None use the 'input_dicom' base
        name.
    write_log: bool (optional, default True)
        If True write the anonimization log.

    Returns
    -------
    output_dicom: str
        the path to the anonimized DICOM file.
    output_log: str
        If 'write_log' is set, the path to the anonimization log.
    """
    # Clean global log
    for key in ANON_LOG.keys():
        ANON_LOG.pop(key)

    # Load the DICOM dataset to anonymize
    if outname is None:
        basedicom = os.path.basename(input_dicom)
        outname = basedicom.split(".")[0]
    else:
        basedicom = outname + ".dcm"
    dataset = dicom.read_file(input_dicom, force=True)

    # Anonymize the dataset
    anonymize_dataset(dataset)

    # Save the anonymized DICOM
    output_dicom = os.path.join(outdir, basedicom)
    dataset.save_as(output_dicom)

    # Save the anonimized log
    output_log = None
    if write_log:
        output_log = os.path.join(outdir, outname + ".json")
        with open(output_log, "w") as open_file:
            json.dump(ANON_LOG, open_file, indent=4)

    return output_dicom, output_log


def anonymize_dataset(dataset, level=1):
    """ Anonymize a pydicom dataset.

    Parameters
    ----------
    dataset: dicom.dataset.Dataset (mandatory)
        a dataset to anonymize.
    """
    # Anonymize the registered tags
    for tag, action in TAGS.values():
        if tag in dataset:
            data_element = dataset[tag]
            tag_repr = repr(data_element.tag)[1:-1].replace(" ", "")
            value = data_element.value
            anon_value = replace_by(value, data_element.VR, action)
            value_repr = repr_dataelement(data_element)
            ANON_LOG.setdefault(tag_repr, []).append((value_repr, anon_value))
            if anon_value is None:
                dataset.pop(data_element.tag)
            else:
                data_element.value = anon_value

    # Anonymize the current dataset applying the registered callbacks
    dataset.walk(callback_main)

    # In supplement 142 the attribute Patient Identity Removed shall be
    # replaced or added to the dataset and the method used for identification
    # need to be specified
    if level == 1:
        add_dataelement(dataset, (0x0012, 0x0062), "YES", "CS")
        add_dataelement(dataset, (0x0012, 0x0063), [
            "Basic Application Confidentiality Profil",
            "Clean Graphics Option",
            "Retain Device Identity Option"], "LO")
        sqdataset = []
        for value, desc in [
                ("113100", "Basic Application Confidentiality Profil"),
                ("113103", "Clean Graphics Option"),
                ("113103", "Retain Device Identity Option")]:
            sqdataset.append((
                ((0x0008, 0x0100), value, "CS"),
                ((0x0008, 0x0104), desc, "LO"),
                ((0x0008, 0x0102), "DCM", "CS")))
        add_dataelement(dataset, (0x0012, 0x0064), sqdataset, "SQ")

    # Check that all tags with 'VR' 'PN' has been anonymized
    # dataset.walk(callback_patient_name)


def callback_main(dataset, data_element):
    """ Called from the dataset 'walk' recursive function, will anonymize
    all DICOM fields by applying all the declared callbacks."""
    # Deal with sequence
    if data_element.VR == "SQ":
        for inner_dataset in data_element.value:
            anonymize_dataset(inner_dataset, level=2)

    # Deal with typed tags
    else:
        for pattern, callback in CALLBACKS.values():

            # Form kwargs
            kwargs = {
                "dataset": dataset,
                "data_element": data_element
            }
            if pattern is not None:
                kwargs["pattern"] = pattern

            # Call the callback
            if callback(**kwargs):
                break



##########################################################################
# NSAp - Copyright (C) CEA, 2013 - 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Package to read/split/convert/anonymize/transcode DICOM files.
"""

from .info import __version__



##########################################################################
# NSAP - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Module that provides DICOM anonymization callbacks.
"""


# Dcmio import
from .utils import replace_by
from .utils import repr_dataelement

# Define global log variable
ANON_LOG = {}
MANUFACTURER = []
CALLBACKS = {}
TAGS = {}
PRIVATE_DEIDENTIFY = {}


def callback_private(dataset, data_element):
    """ Called from the dataset 'walk' recursive function, will anonymize
    all the private fields."""
    # Deal with private tags only
    if data_element.tag.is_private:

        # Check if this private tag needs to be removed
        tag_repr = repr(data_element.tag)[1:-1].replace(" ", "")
        keep_private_tag = False
        for pattern in PRIVATE_DEIDENTIFY.get(MANUFACTURER[0], []):
            if pattern.match(tag_repr):
                keep_private_tag = True
                break

        # Remove tag if requested
        if not keep_private_tag:
            dataset.pop(data_element.tag)
            value = data_element.value
            anon_value = replace_by(value, data_element.VR, "X")
            value_repr = repr_dataelement(data_element)
            ANON_LOG.setdefault(tag_repr, []).append((value_repr, anon_value))
        else:
            print(tag_repr, repr_dataelement(data_element))

        return True

    return False


def callback_tag(dataset, data_element, pattern):
    """ Called from the dataset 'walk' recursive function, will anonymize
    all the tag 'tag'."""
    tag, action = pattern
    if tag in dataset:
        tag_repr = repr(data_element.tag)[1:-1].replace(" ", "")
        value = data_element.value
        anon_value = replace_by(value, data_element.VR, action)
        value_repr = repr_dataelement(data_element)
        ANON_LOG.setdefault(tag_repr, []).append((value_repr, anon_value))
        if anon_value is None:
            dataset.pop(data_element.tag)
        else:
            data_element.value = anon_value

        return True

    return False


def callback_xxxx(dataset, data_element, pattern):
    """ Called from the dataset 'walk' recursive function, will anonymize
    all the '50xx,xxx' fields."""
    tag_repr = repr(data_element.tag)[1:-1].replace(" ", "")

    if pattern.match(tag_repr):
        value = data_element.value
        anon_value = replace_by(value, data_element.VR, "X")
        value_repr = repr_dataelement(data_element)
        ANON_LOG.setdefault(tag_repr, []).append((value_repr, anon_value))
        if anon_value is None:
            dataset.pop(data_element.tag)
        else:
            data_element.value = anon_value

        return True

    return False


def callback_patient_name(dataset, data_element):
    """ Called from the dataset 'walk' recursive function, will set
    a new identitiy to the subject."""
    tag_repr = repr(data_element.tag)[1:-1].replace(" ", "")
    if data_element.VR == "PN":
        if tag_repr not in ANON_LOG:
            raise Exception("Tag '({0})' contains patient information and "
                            "has not been anonymized.".format(tag_repr))



##########################################################################
# NSAp - Copyright (C) CEA, 2013 -2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Module that provides a binding over the 'dcm2nii' command. This module
also defines custom exceptions.
"""



##########################################################################
# NSAP - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Module that provides tools to anonymize DICOM files.
"""


# System import
import dicom
from dicom.dataelem import DataElement_from_raw, RawDataElement


def repr_dataelement(data_element):
    """ Compute the representation of a data element.

    Parameters
    ----------
    data_element: dicom.dataset.DataElement (mandatory)
        a data element to be represented.

    Returns
    -------
    desc: list of str
        the represnetation of the input data element.
    """
    desc = []
    if data_element.VR == "SQ":
        if isinstance(data_element, RawDataElement):
            data_element = DataElement_from_raw(data_element)
        for inner_dataset in data_element.value:
            inner_desc = []
            for inner_tag, innerdata_element in inner_dataset.items():
                inner_repr = repr_dataelement(innerdata_element)
                if isinstance(inner_repr, list):
                    inner_repr = repr(inner_repr)
                inner_desc.append([inner_tag, inner_repr])
            desc.append(inner_desc)
    else:
        desc = repr(data_element.value)
    return desc


def add_dataelement(dataset, tag, value, VR=None):
    """ Update or add a new tag.

    Parameters
    ----------
    dataset: Dataset (mandatory)
        a pydicom DICOM dataset.
    tag: 2-uplet (mandatory)
        a DICOM tag.
    value: str or number (mandatory)
        the actual value.
    VR: str (optional, default None)
        the DICOM value representation: expected if the tag does not exist in
        the current dataset.
    """
    if tag in dataset:
        dataset[tag].value = value
    elif VR is "SQ":
        sequence = []
        for sqvalue in value:
            sqdataset = dicom.dataset.Dataset()
            for sqtag, sqvalue, sqVR in sqvalue:
                add_dataelement(sqdataset, sqtag, sqvalue, sqVR)
            sequence.append(sqdataset)
        element = dicom.dataset.DataElement(tag, VR, sequence)
        dataset.add(element)
    elif VR is not None:
        element = dicom.dataset.DataElement(tag, VR, value)
        dataset.add(element)
    else:
        raise Exception("A VR is expected when the tag to insert does not "
                        "exist in the current dataset.")


def replace_by(value, VR, action,
               default_name="John Doe",
               default_date="18000101",
               default_datetime="180001010000.000000",
               default_time="0000.000000",
               default_text="anon",
               default_code="ANON",
               default_age="000M",
               default_decimal="0.0",
               default_integer="0",
               default_uid="000.000.0"):
    """ Replace a 'value' depending of the input 'action' and the value
    representation 'VR'.

    The following action codes are:

    * D - replace with a non zero length value that may be a dummy value and
      consistent with the VR
    * Z - replace with a zero length value, or a non-zero length value that
      may be a dummy value and consistent with the VR
    * X - remove
    * K - keep (unchanged for non-sequence attributes, cleaned for sequences)
    * C - clean, that is replace with values of similar meaning known not to
      contain identifying information and consistent with the VR
    * U - replace with a non-zero length UID that is internally consistent
      within a set of Instances
    * Z/D - Z unless D is required to maintain IOD conformance (Type 2 versus
      Type 1)
    * X/Z - X unless Z is required to maintain IOD conformance (Type 3 versus
      Type 2)
    * X/D - X unless D is required to maintain IOD conformance (Type 3 versus
      Type 1)
    * X/Z/D - X unless Z or D is required to maintain IOD conformance (Type 3
      versus Type 2 versus Type 1)
    * X/Z/U* - X unless Z or replacement of contained instance UIDs (U) is
      required to maintain IOD conformance (Type 3 versus Type 2 versus Type 1
      sequences containing UID references)

    We use here the PS 3.6 convention.
    """
    if action in ["X", "X/Z", "X/D", "X/Z/D", "X/Z/U*"]:
        return None
    elif action in ["U", "D", "Z", "Z/D"]:
        if VR == "DA":
            return default_date
        elif VR == "AS":
            return default_age
        elif VR == "DS":
            return default_decimal
        elif VR == "DT":
            return default_datetime
        elif VR == "TM":
            return default_time
        elif VR == "FL":
            return float(eval(default_decimal))  # numpy.float32
        elif VR == "FD":
            return float(eval(default_decimal))  # numpy.float64
        elif VR in ["IS"]:
            return default_integer
        elif VR in ["UL", "US", "SS", "SL"]:
            return eval(default_integer)
        elif VR == "PN":
            return default_name
        elif VR == "UI":
            return default_uid
        elif VR == "CS":
            return default_code
        elif VR in ["LO", "LT", "ST", "SH"]:
            return default_text
        elif VR in ["OB", "OW"]:
            return default_integer
        else:
            raise Exception("VR '{0}' is not yet supported. Current value is "
                            "'{1}'.".format(VR, value))
    else:
        raise Exception("Action '{0}' is not yet supported.".format(action))



##########################################################################
# NSAp - Copyright (C) CEA, 2013 - 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Diverse Dicom reading functions to extract precise information.
"""

# System import
import dicom
import os


# Map known request: name: (tag, stack_values)
STANDARD_EXTRACTOR = {
    "get_phase_encoding": ((0x0018, 0x1312), False),
    "get_b_vectors": ((0x0018, 0x9089), True),
    "get_b_values": ((0x0018, 0x9087), True),
    "get_repetition_time": ((0x0018, 0x0080), False),
    "get_date_scan": ((0x0008, 0x0022), False),
    "get_echo_time": ((0x0018, 0x0081), False),
    "get_all_sop_instance_uids": ((0x0008, 0x1155), True),
    "get_sop_storage_type": ((0x0008, 0x0016), False),
    "get_sequence_number": ((0x0020, 0x0011), False),
    "get_nb_slices": ((0x0020, 0x1002), False),
    "get_nb_temporal_position": ((0x0020, 0x0105), False),
    "get_manufacturer_name": ((0x0008, 0x0070), False),
    "get_manufacturer_model_name": ((0x0008, 0x1090), False),
    "get_sequence_name": ((0x0008, 0x103e), False),
    "get_protocol_name": ((0x0018, 0x1030), False),
    "get_serie_instance_uid": ((0x0020, 0x000e), False)
}

# Map to known private tags
PRIVATE_FIELDS = {
    "get_phase_encoding": [],
    "get_b_vectors": [],
    "get_b_values": [],
    "get_repetition_time": [],
    "get_date_scan": [],
    "get_echo_time": [],
    "get_all_sop_instance_uids": [],
    "get_sop_storage_type": [],
    "get_sequence_number": [],
    "get_nb_slices": [(0x2001, 0x102d), (0x2001, 0x1018)],
    "get_nb_temporal_position": [],
    "get_manufacturer_name": [],
    "get_manufacturer_model_name": [],
    "get_sequence_name": [],
    "get_protocol_name": [],
    "get_serie_instance_uid": []
}


def walk(dataset_or_dcmpath, tag, stack_values=False):
    """ Function to extract a tag associated value(s) from a Dicom dataset.

    .. note::

        Recusrive function is required as new enhanced storage presents only
        one Dicom containing several sub-sequence of fields.
        The walked is called on each sub-sequence.

    Parameters
    ----------
    dataset_or_dcmpath: dataset or str (mandatory)
        a pydicom dataset structure or a path to a valid Dicom file.
    tag: 2-uplet (mandatory)
        the Dicom tag of the field containing the value to extract.
    stack_values: bool (optional, default False)
        if set to True, returns all the detected occurences, otherwise the
        first occurence only.

    Returns
    -------
    values: list
        the tag associated value(s). None if the field has not been found in
        the dataset.
    """
    # Deal with input parameters
    if isinstance(dataset_or_dcmpath, str):
        if not os.path.isfile(dataset_or_dcmpath):
            raise ValueError("'{0}' is not a valid Dicom file.".format(
                dataset_or_dcmpath))
        dataset = dicom.read_file(dataset_or_dcmpath, force=True)
    else:
        if not isinstance(dataset_or_dcmpath, dicom.dataset.Dataset):
            raise ValueError("'{0}' is not a 'pydicom' Dataset.".format(
                dataset_or_dcmpath))
        dataset = dataset_or_dcmpath

    # Go through each dataset tags
    values = []

    # Case 1: the desired tag is accessbile in the current dataset level
    if tag in dataset:
        if dataset[tag].VR != "SQ":
            values.append(dataset[tag].value)
            if not stack_values:
                return values
        else:
            sequence = dataset[tag].value
            for sub_dataset in sequence:
                values.extend(
                    walk(dataset=sub_dataset, tag=tag,
                         stack_values=stack_values))
                if not stack_values and len(values) > 0:
                    return values

    # Case 2: go deeper and check tag in Dicom sequences
    for inner_tag in dataset.keys():
        if dataset[inner_tag].VR == "SQ":
            sequence = dataset[inner_tag].value
            for sub_dataset in sequence:
                sub_values = walk(sub_dataset, tag, stack_values=stack_values)
                if sub_values is not None:
                    values.extend(sub_values)
                if not stack_values and len(values) > 0:
                    return values

    # Finally: if no match return None
    if len(values) == 0:
        values = None

    return values


def get_values(dataset_or_dcmpath, extractor):
    """ Get an extractor associated value(s).

    Parameters
    ----------
    dataset_or_dcmpath: dataset or str (mandatory)
        a pydicom dataset structure or a path to a valid Dicom file.
    extractor: str (mandatory)
        the name of the extractor specified in the 'STANDARD_EXTRACTOR'
        global variable.

    Returns
    -------
    values: list or an object
        the phase encoding (none if not found).
    """
    # Deal with input parameters
    if extractor not in STANDARD_EXTRACTOR:
        raise ValueError("'{0}' is not a valid extractor, registered "
                         "extractor are in {1}.".format(
                             extractor, STANDARD_EXTRACTOR.keys()))
    # Get tag associated values
    tag, stack_values = STANDARD_EXTRACTOR[extractor]
    values = walk(dataset_or_dcmpath, tag, stack_values=stack_values)
    if values is None:
        return None
    if not stack_values and stack_values is not None:
        values = values[0]

    # Special cases
    if extractor == "get_repetition_time":
        # > convert in ms
        if values < 1000:
            values *= 1000.

    return values



##########################################################################
# NSAp - Copyright (C) CEA, 2013 - 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Module that provides tools to convert DICOM files.
"""


# System import
from __future__ import print_function
import os
import re
import json
import time

# Third party import
import dicom
import nibabel
import numpy

# Dcmio import
from pydcmio.dcmreader.reader import get_values
from pydcmio.dcmreader.reader import walk
from pydcmio.dcm2nii.wrapper import Dcm2NiiWrapper


def generate_config(niidir, anonymized=True, gzip=True, add_date=True,
                    add_acquisition_number=True, add_protocol_name=True,
                    add_patient_name=True, add_source_filename=True,
                    begin_clip=0, end_clip=0):
    """ Generate a dcm2nii configuration file that disable the interactive
    mode.

    Parameters
    ----------
    niidir: str
        The nifti destination folder.
    anonymized: bool (optional, default True)
        If 'True' then patient name will not be copied to NIfTI header.
    gzip: bool (optional, default True)
        If 'True' then dcm2nii will create compressed .nii.gz files.
    add_date: str (optional, default True)
        If 'True' then dcm2nii will add the requested element in the output
        filename.
    add_acquisition_number: str (optional, default True)
        If 'True' then dcm2nii will add the requested element in the output
        filename.
    add_protocol_name: str (optional, default True)
        If 'True' then dcm2nii will add the requested element in the output
        filename.
    add_patient_name: str (optional, default True)
        If 'True' then dcm2nii will add the requested element in the output
        filename.
    add_source_filename: str (optional, default True)
        If 'True' then dcm2nii will add the requested element in the output
        filename.
    begin_clip: int (optional, default 0)
        Specifies number of volumes to be removed from the beginning of a 4D
        acquisition.
    end_clip: int (optional, default 0)
        Specifies number of volumes to be removed from the end of a 4D
        acquisition.

    Returns
    -------
    config_file: str
        A dcm2nii configuration file written in the input 'niidir' folder.
    """
    # Check the the destination folder exists
    if not os.path.isdir(niidir):
        raise ValueError("'{0}' folder does not exists.".format(niidir))

    # Load the default configuration
    config_file = os.path.join(os.path.dirname(__file__),
                               "dcm2nii_config.json")
    with open(config_file, "r") as open_file:
        config = json.load(open_file)

    # Update the configuration
    config["BOOL"]["Anonymize"] = int(anonymized)
    config["BOOL"]["Gzip"] = int(gzip)
    config["BOOL"]["AppendDate"] = int(add_date)
    config["BOOL"]["AppendAcqSeries"] = int(add_acquisition_number)
    config["BOOL"]["AppendProtocolName"] = int(add_protocol_name)
    config["BOOL"]["AppendPatientName"] = int(add_patient_name)
    config["BOOL"]["AppendFilename"] = int(add_source_filename)
    config["INT"]["BeginClip"] = begin_clip
    config["INT"]["LastClip"] = end_clip
    config["INT"]["OutDirMode"] = 2
    config["STR"]["OutDir"] = niidir

    # Write the configuration file
    config_file = os.path.join(niidir, "dcm2nii.ini")
    with open(config_file, "w") as open_file:
        for section_name, config_items in config.items():
            open_file.write("[{0}]\n".format(section_name))
            for parameter_name, value in config_items.items():
                open_file.write("{0}={1}\n".format(parameter_name, value))

    return config_file


def dcm2nii(input, o, b):
    """ Dicom to nifti conversion using 'dcm2nii'.

    You can specify all the 'dcm2nii' command options as input function
    parameters.

    The basic usage is:
    dcm2nii <options> <sourcenames>

    Options:
    a Anonymize [remove identifying information]: Y,N = Y
    b load settings from specified inifile, e.g. '-b C:\set\t1.ini'
    d Date in filename [filename.dcm -> 20061230122032.nii]: Y,N = N
    e events (series/acq) in filename [filename.dcm -> s002a003.nii]: Y,N = N
    f Source filename [e.g. filename.par -> filename.nii]: Y,N = N
    g gzip output, filename.nii.gz [ignored if '-n n']: Y,N = Y
    i ID  in filename [filename.dcm -> johndoe.nii]: Y,N = N
    n output .nii file [if no, create .hdr/.img pair]: Y,N = Y
    o Output Directory, e.g. 'C:\TEMP' (if unspecified, source directory
    is used)
    p Protocol in filename [filename.dcm -> TFE_T1.nii]: Y,N = Y
    r Reorient image to nearest orthogonal: Y,N
    s SPM2/Analyze not SPM5/NIfTI [ignored if '-n y']: Y,N = N
    t Text report (patient and scan details): Y,N = N
    v Convert every image in the directory: Y,N = Y
    x Reorient and crop 3D NIfTI images: Y,N = N

    Returns
    -------
    files: list of str
        the converted files in nifti format.
    reoriented_files: list of str
        the reoriented converted files in nifti format.
    reoriented_and_cropped_files list of str
        the reoriented and cropped converted files in nifti format.
    bvecs: list of str
        the diffusion directions.
    bvals: list of str
        the diffusion acquisiton b-values.
    """
    # Get the destination folder
    if not os.path.isfile(b):
        raise ValueError("'{0}' is not a valid configuration file.".format(b))
    with open(b, "rt") as open_file:
        lines = open_file.readlines()
    outdirs = [line.replace("OutDir=", "") for line in lines
               if line.startswith("OutDir=")]
    if len(outdirs) != 1:
        raise ValueError("Expect one destination folder ('OutDir=') in "
                         "configuration file '{0}'.".format(b))
    niidir = outdirs[0].rstrip("\n")

    # Call dcm2nii
    dcm2niiprocess = Dcm2NiiWrapper("dcm2nii")
    dcm2niiprocess(cmd=["dcm2nii", "-o", o, "-b", b, input])

    # Format outputs: from nipype
    files = []
    reoriented_files = []
    reoriented_and_cropped_files = []
    bvecs = []
    bvals = []
    skip = False
    last_added_file = None
    for line in dcm2niiprocess.stdout.split("\n"):
        if not skip:
            out_file = None
            # For notgzipped detect files
            if line.startswith("Saving "):
                out_file = line[len("Saving "):]
            # For gzipped outputs files are not absolute
            elif line.startswith("GZip..."):
                out_file = os.path.abspath(
                    os.path.join(niidir, line[len("GZip..."):]))
            # For diffusion
            elif line.startswith("Number of diffusion directions "):
                if last_added_file:
                    base, filename = os.path.split(
                        last_added_file.replace(".gz", "").replace(".nii", ""))
                    bvecs.append(os.path.join(base, filename + ".bvec"))
                    bvals.append(os.path.join(base, filename + ".bval"))
            elif re.search('.*-->(.*)', line):
                val = re.search('.*-->(.*)', line)
                val = val.groups()[0]
                val = os.path.join(niidir, val)
                out_file = val

            if out_file:
                files.append(out_file)
                last_added_file = out_file
                continue

            if line.startswith("Reorienting as "):
                reoriented_files.append(line[len("Reorienting as "):])
                skip = True
                continue
            elif line.startswith("Cropping NIfTI/Analyze image "):
                base, filename = os.path.split(
                    line[len("Cropping NIfTI/Analyze image "):])
                filename = "c" + filename
                reoriented_and_cropped_files.append(
                    os.path.join(base, filename))
                skip = True
                continue
        skip = False

    return files, reoriented_files, reoriented_and_cropped_files, bvecs, bvals


def dcm2niix(input, o, f="%p", z="y", b="y"):
    """ Dicom to nifti conversion using 'dcm2nii'.

    You can specify all the 'dcm2niix' command options as input function
    parameters.

    The basic usage is:
    dcm2niix [options] <in_folder>

    Options:
    b: BIDS sidecar (y/n/o, default n)
    f: filename (%a=antenna  (coil) number, %c=comments, %d=description,
    %e echo number, %f=folder name, %i ID of patient, %m=manufacturer,
    %n=name of patient, %p=protocol, %s=series number, %t=time,
    %u=acquisition number, %z sequence name; default '%f_%p_%t_%s')
    h: show help
    m: merge 2D slices from same series regardless of study time, echo,
    coil, orientation, etc. (y/n, default n)
    o: output directory (omit to save to input folder)
    s: single file mode, do not convert other images in folder
    (y/n, default n)
    t: text notes includes private patient details (y/n, default n)
    v: verbose (y/n, default n)
    x: crop (y/n, default n)
    z: gz compress images (y/i/n, default n) [y=pigz, i=internal, n=no]


    Returns
    -------
    files: list of str
        the converted files in nifti format.
    bvecs: list of str
        the diffusion directions.
    bvals: list of str
        the diffusion acquisiton b-values.
    bids: str
        BIDS sidecar.
    """
    # Call dcm2nii
    dcm2niiprocess = Dcm2NiiWrapper("dcm2niix")
    dcm2niiprocess(cmd=["dcm2niix", "-o", o, "-f", f, "-z", z, "-ba", b,
                        input])

    # Format outputs: from nipype
    files = []
    bvecs = []
    bvals = []
    bids = []
    skip = False
    find_b = False
    for line in dcm2niiprocess.stdout.split("\n"):
        if not skip:
            out_file = None
            if line.startswith("Convert "):
                fname = str(re.search("\S+/\S+", line).group(0))
                output_dir = o
                out_file = os.path.abspath(os.path.join(output_dir, fname))
                # Extract bvals
                if find_b:
                    bvecs.append(out_file + ".bvec")
                    bvals.append(out_file + ".bval")
                    find_b = False
            # Next scan will have bvals/bvecs
            elif "DTI gradients" in line or "DTI gradient directions" in line:
                find_b = True
            else:
                pass
            if out_file:
                files.append(out_file + ".nii.gz")
                if b:
                    bids.append(out_file + ".json")
                continue
        skip = False

    return files, bvecs, bvals, bids


def add_meta_to_nii(nii_file, dicom_dir, dcm_tags, outdir, prefix="f",
                    additional_information=None):
    """ Add dicom tags to Nifti1 image header.

    All selected dicom tag values are set in the 'descrip' Nifti header
    field.

    Parameters
    ----------
    nii_file: str
        The nifti image to fill.
    dicom_dir: str
        The directory containing the dicoms used to generate the nifti image.
        We assume here that this folder contains only Dicom files.
    dcm_tags: list
        A list of 3-uplet of the form (name, tag, stack_values) that will
        be inserted in the 'descrip' Nifti header field. If we want to stack
        all the Dicom dataset values, the 'stack_values' option mist be set to
        True.
    outdir: str
        The destination folder.
    prefix: str (optional, default 'f')
        The output image name prefix.
    additional_information: dict (optional, default None)
        A free dictionary items to be inserted in the 'descrip' image
        header field.

    Returns
    -------
    filled_nii_file: str
        The nifti image with filled header.
    """
    # Set default
    if additional_information is None:
        additional_information = {}

    # Create the destination image path
    if not os.path.isdir(outdir):
        os.makedirs(outdir)

    # Load the first listed dicom image
    dicom_files = os.listdir(dicom_dir)
    dataset = dicom.read_file(os.path.join(dicom_dir, dicom_files[0]),
                              force=True)

    # Load the nifti1 image
    niiimage = nibabel.load(nii_file)

    # Check that we have a nifti1 format image
    filled_nii_file = os.path.join(outdir, prefix + os.path.basename(nii_file))
    if isinstance(niiimage, nibabel.nifti1.Nifti1Image):

        # Fill the nifti1 header
        header = niiimage.get_header()

        # > slice_duration: Time for 1 slice
        repetition_time = get_values(dataset, "get_repetition_time")
        if repetition_time is not None and len(niiimage.shape) > 2:
            repetition_time = float(repetition_time)
            header.set_dim_info(slice=2)
            nb_slices = header.get_n_slices()
            slice_duration = round(repetition_time / nb_slices, 0)
            header.set_slice_duration(slice_duration)

        # > add free dicom fields
        # enhances storage: the value is burried under one or several layer(s)
        # of sequence
        content = {}
        for name, tag, stack_values in dcm_tags:
            content[str(name)] = walk(dataset, tag, stack_values=stack_values)

        # > add/update free content
        content.update(additional_information)

        # Overwrite the 'descrip' header filed
        free_field = numpy.array(json.dumps(content),
                                 dtype=header["descrip"].dtype)
        niiimage.get_header()["descrip"] = free_field

        # Update the image header
        niiimage.update_header()

        # Save the filled image
        nibabel.save(niiimage, filled_nii_file)

    # Unknwon image format
    else:
        raise ValueError(
            "'{0}' is not a Nifti1 image but a '{1}' image.".format(
                nii_file, type(niiimage)))

    return filled_nii_file


def nii2dcm(nii_file, outdir, sid=None, study_id=None, debug=False):
    """ Write a DICOM series from a Nifti array and create appropriate
    meta-data so it can be read by DICOM viewers.

    Writing the 3D image as a DICOM series is done by configuring the
    meta-data dictionary for each of the slices and then writing it in
    DICOM format. In our case we generate all of the meta-data to
    indicate that this series is derived. Note that we write the intensity
    values as is and thus do not set the rescale slope (0028|1053), rescale
    intercept (0028|1052) meta-data dictionary values.

    Parameters
    ----------
    nii_file: str
        The nifti image to convert.
    outdir: str
        The destination folder.
    sid: str, default None
        The subject identifier.
    study_id: str, default None
        The study name.
    debug: bool, default False
        If set try to reload the DICOM serie.

    Returns
    -------
    series_fnames: list of str
        The generated DICOM files.
    """
    import SimpleITK as sitk

    # Read the Nifti image
    img = sitk.ReadImage(nii_file)

    # DICOM does not support directly floating point value. The
    # only thing we can do is apply a well designed linear operation to
    # transform the floating point data to discretized one, here a simple
    # cast.
    print("Casting {0} to 32-bit unsigned integer.".format(
        img.GetPixelIDTypeAsString()))
    img = sitk.Cast(img, sitk.sitkUInt32)

    # Write the 3D image as a serie
    # IMPORTANT:
    # There are many DICOM tags that need to be updated when you modify an
    # original image. This is a delicate opration and requires knowlege of
    # the DICOM standard. This example only modifies some. For a more complete
    # list of tags that need to be modified see:
    #   http://gdcm.sourceforge.net/wiki/index.php/Writing_DICOM
    # If it is critical for your work to generate valid DICOM files,
    # It is recommended to use David Clunie's Dicom3tools to validate the files
    #   (http://www.dclunie.com/dicom3tools.html).
    writer = sitk.ImageFileWriter()

    # Use the study/series/frame of reference information given in the
    # meta-data dictionary and not the automatically generated information
    # from the file IO
    writer.KeepOriginalImageUIDOn()

    # Copy some of the tags and add the relevant tags indicating the change.
    # For the series instance UID (0020|000e), each of the components is a
    # number, cannot start with zero, and separated by a '.' We create a
    # unique series ID using the date and time.
    # Tags of interest:
    modification_time = time.strftime("%H%M%S")
    modification_date = time.strftime("%Y%m%d")
    direction = img.GetDirection()
    series_tag_values = [
        ("0008|0031", modification_time),  # Series Time
        ("0008|0021", modification_date),  # Series Date
        ("0008|0030", modification_time),  # Study Time
        ("0008|0020", modification_date),  # Study Date
        ("0008|0008", "DERIVED\\SECONDARY"),  # Image Type
        ("0010|0020", sid or "NA"),  # Patient ID
        ("0020|0010", study_id or "Convert " + modification_date),  # Study UID
        ("0020|000e", ("1.2.826.0.1.3680043.2.1125." + modification_date +
                       ".1" + modification_time)),  # Series Instance UID
        ("0020|000d", ("1.2.826.0.1.3680043.2.1125." + modification_date +
                       ".1" + modification_time)),  # Study Instance UID
        ("0020|0037", "\\".join(  # Image Orientation (Patient)
            map(str, (direction[0], direction[3], direction[6],
                      direction[1], direction[4], direction[7])))),
        ("0008|103e", "Created-NeuroSpin-SimpleITK")]  # Series Description

    # Write slices to output directory
    list(map(
        lambda index:
            write_slices(series_tag_values, img, index, outdir, writer),
            range(img.GetDepth())))

    # Re-read the series
    # Read the original series. First obtain the series file names using the
    # image series reader.
    series_ids = sitk.ImageSeriesReader.GetGDCMSeriesIDs(outdir)
    if not series_ids:
        raise ValueError("No DICOM files in '{0}'.".format(outdir))
    series_fnames = sitk.ImageSeriesReader.GetGDCMSeriesFileNames(
        outdir, series_ids[0])
    series_reader = sitk.ImageSeriesReader()
    series_reader.SetFileNames(series_fnames)

    if debug:
        # Configure the reader to load all of the DICOM tags (public+private):
        # By default tags are not loaded (saves time).
        # By default if tags are loaded, the private tags are not loaded.
        # We explicitly configure the reader to load tags, including the
        # private ones.
        series_reader.LoadPrivateTagsOn()
        new_img = series_reader.Execute()
        print(new_img.GetSpacing(), "vs", img.GetSpacing())

    return series_fnames


def write_slices(series_tag_values, img, index, outdir, writer):
    """ Write a DICOM slice.

    Parameters
    ----------
    """
    # Get the slice
    image_slice = img[:, :, index]

    # Tags shared by the series.
    list(map(lambda tag_value: image_slice.SetMetaData(
        tag_value[0], tag_value[1]), series_tag_values))

    # Slice specific tags.
    image_slice.SetMetaData(
        "0008|0012", time.strftime("%Y%m%d"))  # Instance Creation Date
    image_slice.SetMetaData(
        "0008|0013", time.strftime("%H%M%S"))  # Instance Creation Time

    # Setting the modality type to MR preserves the slice location.
    image_slice.SetMetaData("0008|0060", "MR")

    # (0020, 0032) image position patient determines the 3D spacing between
    # slices.
    image_slice.SetMetaData("0020|0032", "\\".join(map(
        str, img.TransformIndexToPhysicalPoint((0, 0, index)))))
    image_slice.SetMetaData("0020|0013", str(index))  # Instance Number

    # Write to the output directory and add the extension dcm, to force
    # writing in DICOM format.
    writer.SetFileName(os.path.join(outdir, str(index) + ".dcm"))
    writer.Execute(image_slice)



##########################################################################
# NSAp - Copyright (C) CEA, 2015 - 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module is proposes three main functionalitites:

* a functionality to convert DICOM files to Nifti format.
* a functionality to split DICOM files by serie names.
* a functionality to transcode the subject identifiers.
"""



##########################################################################
# NSAp - Copyright (C) CEA, 2013 - 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Module that provides tools to anonymize DICOM files.
"""


# System import
import os
import json
import random


def transcode_sids(sids, transcoding_table):
    """ Transcode the subject identifiers.

    The transcoded subject identifier is generated randomly (12 digits random
    number between 100000000000 and 999999999999). The procedure checks
    if the subject identifier has already been transcoded.

    Parameters
    ----------
    sids: list of str (mandatory)
        the list of subject identifiers to be transcoded.
    transcoding_table: str (mandatory)
        the transcoding table in JSON format that will be updated if necessary.
    """
    # Load the transcoding table
    if not os.path.isfile(transcoding_table):
        raise ValueError("'{0}' is not a valid transcoding file.".format(
            transcoding_table))
    with open(transcoding_table, "rt") as open_file:
        transcoding = json.load(open_file)

    # Go through each subject id
    for sid in sids:

        # If the original ID is already in the transcoding table, use existing
        # transcodage
        if sid in transcoding:
            continue

        # Otherwise generates a new transcodage randomly
        else:
            transcoded_sid = str(random.randint(100000000000, 999999999999))
            while transcoded_sid in transcoding:
                transcoded_sid = str(
                    random.randint(100000000000, 999999999999))
            transcoding[sid] = transcoded_sid

    # Write the output transcoding table
    with open(transcoding_table, "wt") as open_file:
        json.dump(transcoding, open_file, indent=4)



#! /usr/bin/env python
##########################################################################
# NSAp - Copyright (C) CEA, 2013 - 2016
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

# Current version
version_major = 2
version_minor = 0
version_micro = 2

# Expected by setup.py: string of form "X.Y.Z"
__version__ = "{0}.{1}.{2}".format(version_major, version_minor, version_micro)

# Expected by setup.py: the status of the project
CLASSIFIERS = ["Development Status :: 5 - Production/Stable",
               "Environment :: Console",
               "Environment :: X11 Applications :: Qt",
               "Operating System :: OS Independent",
               "Programming Language :: Python",
               "Topic :: Scientific/Engineering",
               "Topic :: Utilities"]

# Project descriptions
description = """
DICOM managment package.
"""
SUMMARY = """
.. container:: summary-carousel

    pyDcmio is a Python module that can be used to play with DICOM file.
    This package offers:

    * pydcmio_dicomreader: can be use to read specific or registered dicom tags
      recursively.
    * pydcmio_splitseries: can be used to split DICOM files by serie names.
    * pydcmio_dicomanonymizer: can be used to anonymize DICOM files
      (experimental).
    * pydcmio_transcode: can be used to transcode the subject identifiers.
    * pydcmio_dicom2nifti: can be used to convert DICOM files in Nifti format.
"""
long_description = (
    "PYDCMIO: Python DiCoM IO\n\n"
    "[pydcmio] A Python project that provides a wrapping over the 'dcm2nii' "
    "and 'dcm2niix' commands, common tools to read and anonymize DICOM "
    "files, and transcoding features.\n")

# Main setup parameters
NAME = "pyDcmio"
ORGANISATION = "CEA"
MAINTAINER = "Antoine Grigis"
MAINTAINER_EMAIL = "antoine.grigis@cea.fr"
DESCRIPTION = description
LONG_DESCRIPTION = long_description
URL = "https://github.com/neurospin/pydcmio"
DOWNLOAD_URL = "https://github.com/neurospin/pydcmio"
LICENSE = "CeCILL-B"
CLASSIFIERS = CLASSIFIERS
AUTHOR = "pyDcmio developers"
AUTHOR_EMAIL = "antoine.grigis@cea.fr"
PLATFORMS = "OS Independent"
ISRELEASE = True
VERSION = __version__
PROVIDES = ["pydcmio"]
REQUIRES = [
    "numpy>=1.6.1",
    "dicom>=0.9",
    "nibabel>=2.0.2",
    "progressbar2>=3.34.3"
]
EXTRA_REQUIRES = {}
SCRIPTS = [
    "pydcmio/scripts/pydcmio_deface_post",
    "pydcmio/scripts/pydcmio_deface_pre",
    "pydcmio/scripts/pydcmio_dicom2nifti",
    "pydcmio/scripts/pydcmio_dicomanonymizer",
    "pydcmio/scripts/pydcmio_dicomreader",
    "pydcmio/scripts/pydcmio_maskface",
    "pydcmio/scripts/pydcmio_nifti2dicom",
    "pydcmio/scripts/pydcmio_rtstruct",
    "pydcmio/scripts/pydcmio_splitseries",
    "pydcmio/scripts/pydcmio_transcode"
]



##########################################################################
# NSAp - Copyright (C) CEA, 2017
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Module that provides usefull functions to deface an image.
"""

from .mask_face import deface as mask_face
from .mri_deface import deface as mri_deface
from .pdeface import deface as pdeface



#! /usr/bin/env python
##########################################################################
# NSAp - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

# System import
from setuptools import setup, find_packages
import os


release_info = {}
infopath = os.path.join(os.path.dirname(__file__), "pydcmio", "info.py")
with open(infopath) as open_file:
    exec(open_file.read(), release_info)
pkgdata = {
    "pydcmio": ["tests/*.py", "tests/*/*.py"],
    "pydcmio": ["dcmconverter/dcm2nii_config.json"],
    "pydcmio": ["dcmanonymizer/*.json"]
}


# Build the setup
setup(
    name=release_info["NAME"],
    description=release_info["DESCRIPTION"],
    long_description=release_info["LONG_DESCRIPTION"],
    license=release_info["LICENSE"],
    classifiers=release_info["CLASSIFIERS"],
    author=release_info["AUTHOR"],
    author_email=release_info["AUTHOR_EMAIL"],
    version=release_info["VERSION"],
    url=release_info["URL"],
    packages=find_packages(exclude="doc"),
    platforms=release_info["PLATFORMS"],
    extras_require=release_info["EXTRA_REQUIRES"],
    install_requires=release_info["REQUIRES"],
    package_data=pkgdata,
    scripts=release_info["SCRIPTS"]
)



#! /usr/bin/env python
##########################################################################
# NSAp - Copyright (C) CEA, 2013
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Dcm2Nii package wrapper tools.
"""


# System import
import os
import subprocess

# Dcmio import
from .exceptions import Dcm2NiiConfigurationError
from .exceptions import Dcm2NiiRuntimeError

MAP = {
    True: "y",
    False: "n"
}


class Dcm2NiiWrapper(object):
    """ Parent class for the wrapping of Dcm2Nii functions.
    """

    def __init__(self, name):
        """ Initialize the Dcm2NiiWrapper class by setting properly the
        environment.

        Parameters
        ----------
        name: str (mandatory)
            the name of the Dcm2Nii binary to be called.
        """
        self.name = name
        self.cmd = None
        self.environment = os.environ
        self.version = Dcm2NiiWrapper.version(self.name)

    def __call__(self, cmd):
        """ Run the Dcm2Nii command.

        Note that the command is built from the parent frame.
        """
        # Update the command to execute
        self.cmd = [MAP[elem] if isinstance(elem, bool) else elem
                    for elem in cmd]

        # Check Dcm2Nii has been configured so the command can be found
        process = subprocess.Popen(
            ["which", self.cmd[0]],
            env=self.environment,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        self.stdout, self.stderr = process.communicate()
        self.exitcode = process.returncode
        if self.exitcode != 0:
            raise Dcm2NiiConfigurationError(self.cmd[0])

        # Execute the command
        process = subprocess.Popen(
            self.cmd,
            env=self.environment,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        self.stdout, self.stderr = process.communicate()
        self.exitcode = process.returncode

        # Check if the command return a valid exit code
        if self.exitcode != 0:
            raise Dcm2NiiRuntimeError(self.cmd[0], " ".join(self.cmd[1:]),
                                      self.stderr)

    @classmethod
    def version(cls, name):
        """ Get the version of the command.

        Parameters
        ----------
        name: str (mandatory)
            the name of the Dcm2Nii binary to be called.
        """
        # Execute the help command
        process = subprocess.Popen(
            [name],
            env=os.environ,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()
        exitcode = process.returncode

        # Check if the command return a valid exit code
        # TODO: Can't execute this command in parallel
        if exitcode != 0:
            # raise Dcm2NiiRuntimeError(name, "--version", stderr)
            return "uncheckable"

        return stdout.splitlines()[0]



##########################################################################
# NSAp - Copyright (C) CEA, 2017
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
Defacing with the Poldrack, R. 'pydeface' Python module.
"""

# System import
import os
import warnings

# Pyconnectome
try:
    from pyconnectome import DEFAULT_FSL_PATH
    from pyconnectome.wrapper import FSLWrapper
except:
    DEFAULT_FSL_PATH = ""
    warnings.warn("PyConnectome is not installed.")


def deface(input_files, outdir, reference_file=None, fsl_sh=DEFAULT_FSL_PATH):
    """ Deface MRI head images using Poldrack, R. 'pydeface' Python module.

    Parameters
    ----------
    input_files: list of str
        Input MRI head images to be defaced.
    outdir: str
        The output folder.
    reference_file: str (optional, default None)
        The image that must be used as reference if more than one image
        have been supplied as input.
    fsl_sh: str (optional, default DEFAULT_FSL_PATH)
        The FSL configuration file.

    Returns
    -------
    deface_files: list of str
        The defaced input MRI head images.
    snap_files: list of str
        The corresponding snaps that can be used to check the defacing result.
    """
    # Check input parameters
    outdir = os.path.abspath(outdir)
    if len(input_files) == 0:
        raise ValueError("You must specify at least one image.")
    elif len(input_files) > 1:
        raise ValueError("Mutliple input files not yet supported.")
        # if reference_file is None or reference_file not in input_files:
        #     raise ValueError("If more than one image is specified, you must "
        #                      "also specify one of them as reference. The "
        #                      "reference image will be used for spatial "
        #                      "co-registration with an atlas, and others "
        #                      "will use the reference facial mask.")

    # Link the input files
    local_input_files = []
    for fp in input_files:
        local_input_files.append(os.path.join(outdir, os.path.basename(fp)))
        if not os.path.exists(local_input_files[-1]):
            os.symlink(fp, local_input_files[-1])
        else:
            warnings.warn(
                "'{0}' file already here.".format(local_input_files[-1]))

    # Define the command
    deface_file = os.path.join(
        outdir, "pydeface_" + os.path.basename(input_files[0]))
    cmd = ["pydeface.py", local_input_files[0], deface_file]

    # Call defacing
    wrapper = FSLWrapper(cmd, shfile=fsl_sh, env=os.environ)
    wrapper()

    # Remove links
    for fp in local_input_files:
        os.remove(fp)

    return [deface_file], None
